from .btree_to_list import btree_to_list
from .btree_list_printer import display_btree
from .btree_caster import cast
from .btree_audit import audit

def print_btree(
    root,
    val='val',
    left='left',
    right='right'
) -> None:
    """
    Args:
        root: root of binary tree

        val: name of .val variable in Node
        left: name of .left variable in Node
        right: name of .right variable in Node

            eg, if your Node's class definition is something like:

            class BNode:
                def __init__(self, v):
                    self.v = v
                    self.l = None
                    self.r = None

            use print_btree(root, val='v', left='l', right='r')

    """
    root = cast(root, val=val, left=left, right=right)
    ls = btree_to_list(root)
    rows = display_btree(ls)
    rows = audit(rows)

    print();[print(row) for row in rows[::-1]];print()

__all__ = ['print_btree']


from collections import deque

def get_height(root):
    """
    given root of btree, find height of tree
    """
    largest = -1
    queue = deque([(root, 1)])
    while queue:
        current, lvl = queue.popleft()
        if lvl > largest:
            largest = lvl
        if current.left: queue.append((current.left, lvl+1))
        if current.right: queue.append((current.right, lvl+1))
    return largest

def btree_to_list(root) -> list[list]:
    """
    generates 2d list representing btree
        eg. [[1], [2,3], [4,5,6,7]]
    """
    height = get_height(root)
    out = [[] for i in range(height)]
    queue = deque([(root, 0)])
    while queue:
        current, lvl = queue.popleft()
        if lvl > height-1:
            continue
        
        if current is None:
            out[lvl].append(None)
            queue.append((None, lvl+1))
            queue.append((None, lvl+1))
        else:
            out[lvl].append(current.val)
            queue.append((current.left, lvl+1))
            queue.append((current.right, lvl+1))
    return out


from src.print_btree import print_btree
from src.print_btree.utils import BTree



class N:
    def __init__(self, val):
        self.val = val
        self.l = self.r = None

def show(root): print_btree(root, left='l', right='r')

# root = N(1)
# root.l = N(2)
# root.r = N(3)
# root.l.l = N(4)
# root.l.l.r = N(5)
# root.l.l.r.l = N(6)
# show(root)

root = N('apple')
root.l = N(1)
root.r = N(2)
root.l.l = N('orange')
root.l.r = N('pear')
root.r.r = N('pie')
root.l.r.l = N(100)
root.l.r.r = N(200)
root.r.r.l = N('hello')
show(root)




from src.print_btree.utils import BTree, Node
from src.print_btree.btree_to_list import get_height, btree_to_list

def test_get_height():
    root = BTree.gen_btree([1])
    assert get_height(root) == 1

    root = BTree.gen_btree([1,2])
    assert get_height(root) == 2

    root = BTree.gen_btree([1,2,3])
    assert get_height(root) == 2

    root = BTree.gen_btree([1,2,3,4])
    assert get_height(root) == 3

    root = BTree.gen_btree([1,2,3,4,5,6,7])
    assert get_height(root) == 3

    root = BTree.gen_btree([1,2,3,4,5,6,7,8])
    assert get_height(root) == 4

def test_btree_to_list():
    root = BTree.gen_btree([1,2,3,4,5,6,7])
    ls = btree_to_list(root)
    assert ls == [[1], [2,3], [4,5,6,7]]

    root = Node(1)
    root.left = Node(2)
    root.right = Node(3)
    root.left.left = Node(5)
    root.right.left = Node(6)
    ls = btree_to_list(root)
    assert ls == [[1], [2,3], [5,None,6,None]]

    root = Node(1)
    root.left = Node(2)
    root.right = Node(3)
    root.left.left = Node(5)
    root.right.left = Node(6)
    root.right.left.right = Node('apple')
    ls = btree_to_list(root)
    assert ls == [[1], [2,3], [5,None,6,None], [None, None, None, None, None, 'apple', None, None]]


class StandardizedNode:
    def __init__(self, val):
        self.val = val
        self.left = self.right = None

def cast(node, val='val', left='left', right='right'):
    if node is None: return None
    node_val = getattr(node, val)
    node_val = str(node_val).replace(' ', '-')
    s_node = StandardizedNode(node_val)

    s_node.left = cast(getattr(node, left), val=val, left=left, right=right)
    s_node.right = cast(getattr(node, right), val=val, left=left, right=right)
    return s_node




from collections import deque

class Node:
    def __init__(self, val):
        self.val = val
        self.left = None
        self.right = None

    def __str__(self) -> str:
        val = self.val
        left = right = None
        if self.left: left = left.val
        if self.right: right = right.val
        return f'Node({val=}, {left=}, {right=})'

class BTree:
    @staticmethod
    def gen_btree(ls: list) -> Node:
        """
        ls: a list of values eg. [1,2,3,4,5,6,7]

        output: root of a binary tree, with values inserted in order
        """
        root = Node(ls[0])
        for i in range(1, len(ls)):
            BTree.add_one(root, ls[i])
        return root

    @staticmethod
    def add_one(node: Node, val) -> None:
        """
        add new val to node at next available spot
        """
        queue = deque([node])
        while queue:
            current = queue.popleft()

            if current.left is None: 
                current.left = Node(val)
                return
            if current.right is None:
                current.right = Node(val)
                return
            
            queue.append(current.left)
            queue.append(current.right)


if __name__ == '__main__':
    root = BTree.gen_btree([1,2,3,4,5,6,7])

    print(root.left.left)




from src.print_btree.utils import BTree

def test_gen_btree():
    root = BTree.gen_btree([1, 2, 3])
    assert root.val == 1
    assert root.left.val == 2
    assert root.right.val == 3
    assert root.left.left is None
    assert root.left.right is None
    assert root.right.left is None
    assert root.right.right is None

def test_gen_btree_2():
    root = BTree.gen_btree([1,2,3,4,5,6,7,8])
    assert root.val == 1
    assert root.left.val == 2
    assert root.right.val == 3
    assert root.left.left.val == 4
    assert root.left.right.val == 5
    assert root.right.left.val == 6
    assert root.right.right.val == 7
    assert root.left.left.left.val == 8


def rid_pipes(top, bottom):
    """
    removes unneccesary pipes

    inputs:
        '|  |  |  |  |  |  |  |'    # bottom
        '   9           8     7'    # top

    returns:
        '   |           |     |'
    """
    out = ''
    for t,b in zip(top, bottom):
        if t == ' ': out += ' '
        else: out += b
    return out

def rid_underscores(top, bottom):
    """
    removes unnecessary underscores

    inputs:
        ' _4__  _ __  _6__  _7__'   # bottom
        '    |           |     |'   # top

    returns:
        '  4__         6__   7__'
    """
    out = bottom
    l = r = -1
    bottom += ' '
    for i, ch in enumerate(bottom):
        if ch == '_':
            if l < 0: l = i
            r = i

        else:
            # if l >= 0:
            #     if '4' in bottom:
            #         print(l,r,[bottom[l:r+1], top[l:r+1]])

            if l >= 0 and ('|' not in top[l:r+1]):
                out = out[:l] + ' '*(r-l+1) + out[r+1:]
            l = r = -1
    return out

b = '_4__  _ __  _6__  _7__'   # bottom
t = '   |           |     |'   # top

# print(rid_underscores(t, b))


def remove_front_spaces(rows: list[str]) -> list[str]:
    """
    remove common whitespace at the front
    input:
        [
            '     aa',
            '       aa',
            '      aa',
        ]
    
    output:
        [
            'aa',
            '  aa',
            ' aa',
        ]
    """
    smallest = float('inf')
    for row in rows:
        num_spaces = len(row) - len(row.lstrip())
        if num_spaces < smallest:
            smallest = num_spaces
    for i, row in enumerate(rows):
        rows[i] = row[smallest:]
    return rows



def audit(rows: list[str]):
    rows[0] = rows[0].replace('?', ' ')
    for i in range(len(rows)-1):
        # change bottom based on top
        top = rows[i]
        bottom = rows[i+1]
        
        if i%2 == 1:
            # top contains pipes, bottom contains values
            bottom = bottom.replace('?', ' ')
            rows[i+1] = rid_underscores(top, bottom)
        else:
            # bottom contains pipes, top contains values:=
            rows[i+1] = rid_pipes(top, bottom)

    rows = remove_front_spaces(rows)
    return rows



import os
import re

with open('pyproject.toml') as f:
    text = f.read()

old_version = re.findall('version = "(.*?)"', text)[0]
a, b, c = old_version.split('.')
x = int(c) + 1
new_version = f'{a}.{b}.{x}'
text = re.sub(old_version, new_version, text)

with open('pyproject.toml', 'w') as f:
    f.write(text)    

commands = [
    'rm -rf ./dist',
    'python3 -m build',
    'python3 -m twine upload dist/*'
]

for command in commands:
    print(command)
    os.system(command)


class BTreeSpacingException(Exception):
    pass

def get_pipes(string: str) -> str:
    '''
    given a row, get pipes at correct positions

    'apple      orange         pear        pineapple'
                        to
    '  |          |             |              |'
    '''
    out = ''
    current = ''
    for c in string:
        if c == ' ': 
            if current: 
                out += '|'.center(len(current))
                current = ''
            out += ' '
        else:
            current += c
    if current: 
        out += '|'.center(len(current))
    return out.rstrip()

def get_row(
    string: str, 
    ls: list
) -> str:
    '''
    inputs:
        string: string of pipes
        ls: list of values to insert

    given inputs, return next row

    '  |    |    |      |    |      |     |    |'
                        to
    '  __a___    ___b____    ___c____     __d___'   
    '''
    indexes = []
    for i,c in enumerate(string):
        if c == '|': indexes.append(i)
    out = ' ' * indexes[0]
    ls_i = 0
    for i in range(len(indexes)-1):
        l = indexes[i]
        r = indexes[i+1]
        if i % 2 == 0:
            val = str(ls[ls_i] or '?').center(r-l+1, '_')

            if val[0] != '_' or val[-1] != '_':
                raise BTreeSpacingException()

            out += val
            ls_i += 1
        else:
            out += ' ' * (r-l-1)
    return out


def display_btree_attempt(
    ls: list[list],
    num_spaces
) -> None:
    """
    inputs:
        ls: 2d list representing btree eg [[1], [2,3], [4,5,6,7]]
        num_spaces: num spaces between values in bottom of tree
    """
    ls = ls[::-1]
    def get_base(base, num_spaces):
        out = ''
        for i,val in enumerate(base):
            out += str(val or '?') + ' '*num_spaces
        return out.strip()

    out = [get_base(ls[0], num_spaces=num_spaces)]
    for row in ls[1:]:
        pipes = get_pipes(out[-1])
        out.append(pipes)

        row = get_row(out[-1], row)
        out.append(row)

    return out


def display_btree(ls: list[list]) -> None:
    """
    keep trying to display btree until BTreeSpacingException is not raised

    inputs:
        ls: 2d list representing btree eg [[1], [2,3], [4,5,6,7]]
    """
    num_spaces = 1
    while True:
        try:
            return display_btree_attempt(ls, num_spaces)
        except BTreeSpacingException as e:
            num_spaces += 1
        
